# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for extracting meta data about YouTube channels and videos."""

from __future__ import annotations

import collections
import dataclasses
import os
from typing import Any, Literal

from googleapiclient.discovery import build

from googleads_housekeeper.domain.external_parsers import base_parser


class YouTubeDataConnector:
  """Specifies connections to YouTube.

  Attributes:
      service: Build YouTube service.
  """

  def __init__(
    self,
    api_version: str = 'v3',
    developer_key: str = os.getenv('YOUTUBE_DATA_API_KEY'),
  ):
    """Initializes YouTubeDataConnector.

    Args:
        api_version: Version of YouTube Data API.
        developer_key: API key to access YouTube Data API.

    Returns:
        Description of return.
    """
    self.service = build('youtube', api_version, developerKey=developer_key)

  def get_response(
    self,
    type_: Literal['videos', 'channels'],
    elements: str,
    placements: collections.abc.Sequence[str],
  ):
    """Gets API response for a correct type.

    Args:
        type_: Service type.
        elements: Elements to return in response.
        placements: Identifiers for videos/channels.

    Returns:
        Description of return.
    """
    if type_ == 'videos':
      service = self.service.videos()
    elif type_ == 'channels':
      service = self.service.channels()
    else:
      raise ValueError(f'Unsupported resource {type_}')
    return service.list(part=elements, id=placements).execute()


@dataclasses.dataclass
class ChannelInfo(base_parser.EntityInfo):
  """Contains meta information of YouTube Channel.

  Attributes:
      placement: Id of YouTube channel.
      title: Title of the channel.
      description: Channel description.
      viewCount: Number of channel views.
      subscriberCount: Number of channel subscribers.
      videoCount: Number of videos on the channel.
      topicCategories: Categories associated with the channel.
      is_processed: Whether YouTube channel was successfully parsed.
  """

  placement: str
  title: str | None = None
  description: str | None = None
  country: str | None = None
  viewCount: int = 0
  subscriberCount: int = 0
  videoCount: int = 0
  topicCategories: str = ''
  is_processed: bool = True


@dataclasses.dataclass
class VideoInfo(base_parser.EntityInfo):
  """Contains meta information of YouTube Video.

  Attributes:
      placement: Id of YouTube video.
      title: Title of the video.
      description: Channel description.
      defaultLanguage: Language of the video.
      defaultAudioLanguage: Audio language of the video.
      commentCount: Number of video comments.
      favouriteCount: Number of times video is added to favorites.
      likeCount: Number of video likes.
      viewCount: Number of videos views.
      madeForKids: Whether is the video is made for kids.
      tags: Tags associated with the video.
      topicCategories: Categories associated with the video.
      is_processed: Whether YouTube video was successfully parsed.
  """

  placement: str
  title: str | None = None
  description: str | None = None
  defaultLanguage: str | None = None
  defaultAudioLanguage: str | None = None
  commentCount: int = 0
  favouriteCount: int = 0
  likeCount: int = 0
  viewCount: int = 0
  madeForKids: bool = False
  tags: str = ''
  topicCategories: str = ''
  is_processed: bool = True


class ChannelInfoParser(base_parser.BaseParser):
  """Performs YouTube channel parsing."""

  def __init__(
    self, data_connector: type[YouTubeDataConnector] = YouTubeDataConnector
  ):
    """Initializes ChannelInfoParser.

    Args:
        data_connector: Connector to get data from YouTube.
    """
    self.data_connector = data_connector

  def parse(
    self, placements: collections.abc.Sequence[str]
  ) -> list[ChannelInfo]:
    """Parses provided YouTube channels.

    Args:
        placements: Sequence of YouTube channel ids.

    Returns:
        Parsed information in ChannelInfo format.
    """
    response = self.data_connector().get_response(
      'channels', 'id,snippet,statistics,topicDetails', placements
    )
    if not (items := response.get('items')):
      return [
        ChannelInfo(placement=channel_id, is_processed=False)
        for channel_id in placements
      ]
    results: list[ChannelInfo] = []
    for item in items:
      if snippet := item.get('snippet'):
        title = snippet.get('title')
        description = snippet.get('description')
        country = snippet.get('country')
      else:
        title = None
        description = None
        country = None
      if statistics := item.get('statistics'):
        subscriberCount = safe_cast(int, statistics.get('subscriberCount'))
        viewCount = safe_cast(int, statistics.get('viewCount'))
        videoCount = safe_cast(int, statistics.get('videoCount'))
      else:
        subscriberCount = 0
        viewCount = 0
        videoCount = 0
      topics = parse_topic_details(item.get('topicDetails'))
      results.append(
        ChannelInfo(
          placement=item.get('id'),
          title=title,
          description=description,
          country=country,
          viewCount=viewCount,
          subscriberCount=subscriberCount,
          videoCount=videoCount,
          topicCategories=topics,
        )
      )
    return results


class VideoInfoParser(base_parser.BaseParser):
  """Performs YouTube video parsing."""

  def __init__(
    self, data_connector: type[YouTubeDataConnector] = YouTubeDataConnector
  ):
    """Initializes VideoInfoParser.

    Args:
        data_connector: Connector to get data from YouTube.
    """
    self.data_connector = data_connector

  def parse(self, placements: collections.abc.Sequence[str]) -> list[VideoInfo]:
    """Parses provided YouTube videos.

    Args:
        placements: Sequence of YouTube video ids.

    Returns:
        Parsed information in VideoInfo format.
    """
    response = self.data_connector().get_response(
      'videos',
      'id,status,snippet,statistics,contentDetails,topicDetails',
      placements,
    )
    if not (items := response.get('items')):
      return [
        VideoInfo(placement=video_id, is_processed=False)
        for video_id in placements
      ]
    results: list[VideoInfo] = []
    for item in items:
      if snippet := item.get('snippet'):
        title = snippet.get('title')
        description = snippet.get('description')
        defaultLanguage = snippet.get('defaultLanguage')
        defaultAudioLanguage = snippet.get('defaultAudioLanguage')
        tags = snippet.get('tags')
      else:
        title = None
        description = None
        defaultLanguage = None
        defaultAudioLanguage = None
        tags = ''
      if statistics := item.get('statistics'):
        commentCount = safe_cast(int, statistics.get('commentCount'))
        favouriteCount = safe_cast(int, statistics.get('favouriteCount'))
        likeCount = safe_cast(int, statistics.get('likeCount'))
        viewCount = safe_cast(int, statistics.get('viewCount'))
      else:
        commentCount = 0
        favouriteCount = 0
        likeCount = 0
        viewCount = 0
      if status := item.get('status'):
        madeForKids = safe_cast(bool, status.get('madeForKids'))
      else:
        madeForKids = False
      topics = parse_topic_details(item.get('topicDetails'))
      if tags:
        tags = ','.join(tags)
      results.append(
        VideoInfo(
          placement=item.get('id'),
          title=title,
          description=description,
          defaultLanguage=defaultLanguage,
          defaultAudioLanguage=defaultAudioLanguage,
          commentCount=commentCount,
          favouriteCount=favouriteCount,
          likeCount=likeCount,
          viewCount=viewCount,
          madeForKids=madeForKids,
          tags=tags,
          topicCategories=topics,
        )
      )
    return results


def parse_topic_details(
  topic_details: collections.abc.Mapping[str, Any] | None,
) -> str:
  """Converts Wikipedia URLs to topic names."""
  if not topic_details:
    return ''
  if not (topic_categories := topic_details.get('topicCategories')):
    return ''
  return ','.join(list({topic.split('/')[-1] for topic in topic_categories}))


def safe_cast(callable_, value: str | None):
  if value:
    return callable_(value)
  return callable_()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handles interacting with Google Ads API."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

import gaarf
import numpy as np
import pandas as pd

from googleads_housekeeper.domain.core import exclusion_specification, task
from googleads_housekeeper.domain.placement_handler import entities
from googleads_housekeeper.services import enums


class PlacementFetcher:
  """Handles fetching of placement data from Google Ads API.

  Attributes:
    report_fetcher: An instance of AdsReportFetcher to handle API requests.
  """

  def __init__(
    self, report_fetcher: gaarf.query_executor.AdsReportFetcher
  ) -> None:
    """Initializes PlacementFetcher.

    Args:
      report_fetcher: An instance of AdsReportFetcher to handle API requests.
    """
    self.report_fetcher = report_fetcher

  def get_placements_for_account(
    self,
    account: str,
    task_obj: task.Task,
    mode: Literal['preview', 'action'] = 'action',
  ) -> gaarf.report.GaarfReport | None:
    """Gets placement data from a given account based on task definition.

    Args:
      account: Google Ads account to get data from.
      task_obj: A task defining parameters of fetching.
      mode: Whether to fetch all data or only a subset.

    Returns:
      Report with fetched placement data.
    """
    specification = (
      exclusion_specification.ExclusionSpecification.from_expression(
        task_obj.exclusion_rule
      )
    )
    runtime_options = specification.define_runtime_options()
    if not (
      placements := self._get_placement_performance_data(
        account, task_obj, mode
      )
    ):
      return None
    if runtime_options.is_conversion_query:
      if (
        placements_by_conversion_name
        := self._get_placement_conversion_split_data(account, task_obj)
      ):
        conversion_split_exclusion_specification = (
          exclusion_specification.ExclusionSpecification(
            specifications=[runtime_options.conversion_rules]
          )
        )
        placements_by_conversion_name = (
          conversion_split_exclusion_specification.apply_specifications(
            placements_by_conversion_name
          )
        )
        placements = self._join_conversion_split(
          placements,
          placements_by_conversion_name,
          runtime_options.conversion_name,
        )
      else:
        self._extend_placement_with_conversion_split_columns(
          placements=placements
        )
    return self._aggregate_placements(placements, task_obj.exclusion_level)

  def _get_placement_performance_data(
    self, account: str, task_obj: task.Task, mode: Literal['preview', 'action']
  ) -> gaarf.report.GaarfReport:
    """Gets placement performance data (clicks, impressions, etc.).

    Args:
      account: Google Ads account to get data from.
      task_obj: A task defining parameters of fetching.

    Returns:
      Report with fetched placement data.
    """
    placements = self._get_placement_data(
      task_obj=task_obj, account=account, mode=mode
    )
    if 'YOUTUBE_VIDEO' in task_obj.placement_types:
      if detail_placements := self._get_placement_data(
        task_obj=task_obj,
        account=account,
        placement_level_granularity='detail_placement_view',
        mode=mode,
      ):
        placements = placements + detail_placements
    return placements

  def _get_placement_conversion_split_data(
    self, account: str, task_obj: task.Task
  ) -> gaarf.report.GaarfReport:
    """Gets placement conversion split data (conversion_name, conversions).

    Args:
      account: Google Ads account to get data from.
      task_obj: A task defining parameters of fetching.

    Returns:
      Report with fetched placement conversion split data.
    """
    placements = self._get_placement_data(
      task_obj=task_obj,
      account=account,
      query_class=entities.PlacementsConversionSplit,
    )
    if 'YOUTUBE_VIDEO' in task_obj.placement_types:
      if detail_placements := self._get_placement_data(
        task_obj=task_obj,
        account=account,
        query_class=entities.PlacementsConversionSplit,
      ):
        placements = placements + detail_placements
    return placements

  def get_already_excluded_placements(
    self, account: str, exclusion_level: enums.ExclusionLevelEnum
  ) -> dict[str, list[str]]:
    """Fetches negative placement criteria from Google Ads account.

    Negative criteria are fetched based on specified exclusion level (
    ACCOUNT, CAMPAIGN, AD_GROUP).

    Args:
      account: Google Ads account to get data from.
      exclusion_level: Level of exclusion (ACCOUNT, CAMPAIGN, AD_GROUP).

    Returns:
      Mapping between entity_id from specified exclusion level to all
      negative criteria ids.
    """
    already_excluded_placements = self.report_fetcher.fetch(
      entities.AlreadyExcludedPlacements(exclusion_level), account
    )
    if not already_excluded_placements:
      return {}
    for row in already_excluded_placements:
      row['placement'] = (
        row.website_url or row.app_id or row.video_id or row.channel_id or ''
      )
    return already_excluded_placements.to_dict(
      key_column='level_id',
      value_column='placement',
      value_column_output='list',
    )

  def get_placement_exclusion_lists(
    self, customer_ids: Sequence[int]
  ) -> dict[str, str]:
    placement_exclusion_lists = self.report_fetcher.fetch(
      entities.PlacementExclusionLists(), customer_ids
    )
    return placement_exclusion_lists.to_dict(
      key_column='name',
      value_column='resource_name',
      value_column_output='scalar',
    )

  def _build(self, spec) -> tuple[str, str] | None:
    if spec.name in (
      'clicks',
      'impressions',
      'cost',
      'conversions',
      'conversions_value',
    ):
      value = int(int(spec.value) * 1e6) if spec.name == 'cost' else spec.value
      name = f'{spec.name}_micros' if spec.name == 'cost' else spec.name
      return (
        f'metrics.{name} {spec.operator} {value}',
        self._build_metric(spec.name),
      )
    return None

  def _build_metric(self, metric_name: str) -> str:
    name = (
      f'{metric_name}_micros / 1e6' if metric_name == 'cost' else metric_name
    )
    return f'metrics.{name} AS {metric_name}'

  def _get_placement_data(
    self,
    task_obj: task.Task,
    account: str,
    query_class: gaarf.base_query.BaseQuery = entities.Placements,
    placement_level_granularity: str = 'group_placement_view',
    mode: Literal['preview', 'action'] = 'action',
  ) -> gaarf.report.Report:
    """Helper method for building query and fetching data from Ads API.

    Args:
      task_obj: A Task that contains necessary data for building query.
      account: Google Ads account to fetch data from.
      query_class: Class defining which query will be built.
      placement_level_granularity: Resource name to get data from.

    Returns:
      Report containing placement performance data.
    """
    spec = exclusion_specification.ExclusionSpecification.from_expression(
      task_obj.exclusion_rule
    )
    ads_specs = spec.ads_specs_entries.specifications
    metrics: set[str] = set()
    filters: set[str] = set()
    mappings = {
      'avg_cpc': ['clicks', 'cost'],
      'cost_per_conversion': ['cost', 'conversions'],
      'cost_per_all_conversion': ['cost', 'all_conversions'],
      'ctr': ['clicks', 'impressions'],
      'avg_cpv': ['cost', 'video_views'],
      'avg_cpm': ['cost', 'impressions'],
    }
    for specs in ads_specs:
      for spec in specs:
        if compound_metrics := mappings.get(spec.name):
          for metric in compound_metrics:
            metrics.add(self._build_metric(metric))
        else:
          ads_filter, ads_metric = self._build(spec)
          metrics.add(ads_metric)
          filters.add(ads_filter)

    limit = 10_000 if mode == 'preview' else 0
    placement_query = query_class(
      placement_types=task_obj.placement_types,
      placement_level_granularity=placement_level_granularity,
      start_date=task_obj.start_date,
      end_date=task_obj.end_date,
      limit=limit,
      metrics=metrics,
      filters=filters,
    )
    return self.report_fetcher.fetch(placement_query, customer_ids=account)

  def _join_conversion_split(
    self,
    placements: gaarf.report.GaarfReport,
    placements_by_conversion_name: gaarf.report.GaarfReport,
    conversion_name: str,
  ) -> gaarf.report.GaarfReport:
    """Joins placements performance data with its conversion split data.

    Args:
      placements:
        Report with placement performance data.
      placements_by_conversion_name:
        Report with placements conversion split data.
      conversion_name:
        Conversion_name(s) that should be used to create a dedicated
        column in joined report.

    Returns:
      New report with extra conversion specific columns.
    """
    placements_by_conversion_name = placements_by_conversion_name.to_pandas()
    final_report_values = []
    for performance_data_row in placements:
      conversions, all_conversions = self._extract_conversion_split_metrics(
        placements_by_conversion_name,
        performance_data_row.ad_group_id,
        performance_data_row.placement,
        performance_data_row.placement_type,
      )
      final_report_values.append(
        performance_data_row.data
        + [conversion_name, conversions, all_conversions]
      )
    columns = placements.column_names + [
      'conversion_name',
      'custom_conversions',
      'custom_all_conversions',
    ]
    return gaarf.report.GaarfReport(
      results=final_report_values, column_names=columns
    )

  def _extract_conversion_split_metrics(
    self,
    placements_by_conversion_name: pd.DataFrame,
    ad_group_id: int,
    placement: str,
    placement_type: str,
  ) -> tuple[float, float]:
    """Checks for rows with provided ad_group_id, placement, placement_type.

    Args:
      placements_by_conversion_name: DataFrame with conversion split data.
      ad_group_id: Ad group id.
      placement: Name of placement.
      placement_type: Type of placement.

    Returns:
      Conversions and all_conversions satisfying the condition; if nothing is
        found provides 0 as a placeholder.
    """
    conversion_row_filter = (
      (placements_by_conversion_name.ad_group_id == ad_group_id)
      & (placements_by_conversion_name.placement == placement)
      & (placements_by_conversion_name.placement_type == placement_type)
    )
    filtered_conversion_rows = placements_by_conversion_name.loc[
      conversion_row_filter
    ]
    if not (conversions := sum(filtered_conversion_rows['conversions'].values)):
      conversions = 0.0
    if not (
      all_conversions := sum(filtered_conversion_rows['all_conversions'].values)
    ):
      all_conversions = 0.0
    return conversions, all_conversions

  def _extend_placement_with_conversion_split_columns(
    self, placements: gaarf.report.GaarfReport
  ) -> None:
    """Extend placements performance data with conversion split columns.

    Args:
      placements: Report with placement performance data.
    """
    for row in placements:
      row['conversion_name'] = ''
      row['custom_conversions'] = 0.0
      row['custom_all_conversions'] = 0.0

  def _aggregate_placements(
    self,
    placements: gaarf.report.GaarfReport,
    exclusion_level: str | enums.ExclusionLevelEnum,
    perform_relative_aggregations: bool = True,
  ) -> gaarf.report.GaarfReport:
    """Aggregates placements to a desired exclusion_level.

    By default Placements report returned on Ad Group level, however exclusion
    can be performed on Campaign, Account and MCC level. By aggregating report
    to a desired level exclusion specification can be property applied to
    identify placements that should be excluded.

    Args:
      placements:
        Report with placement related metrics.
      exclusion_level:
        Desired level of aggregation.
      perform_relative_aggregations:
        Whether or not calculate relative metrics (CTR, CPC, etc.)

    Returns:
      Updated report aggregated to desired exclusion level.
    """
    if not isinstance(exclusion_level, enums.ExclusionLevelEnum):
      exclusion_level = getattr(enums.ExclusionLevelEnum, exclusion_level)
    base_groupby = [
      'placement',
      'placement_type',
      'name',
      'criterion_id',
      'url',
    ]
    aggregation_dict = dict.fromkeys(
      [
        'clicks',
        'impressions',
        'cost',
        'conversions',
        'video_views',
        'interactions',
        'all_conversions',
        'view_through_conversions',
      ],
      'sum',
    )
    relative_aggregations_dict = {
      'ctr': ['clicks', 'impressions'],
      'avg_cpc': ['cost', 'clicks'],
      'avg_cpm': ['cost', 'impressions'],
      'avg_cpv': ['cost', 'video_views'],
      'video_view_rate': ['video_views', 'impressions'],
      'interaction_rate': ['interactions', 'clicks'],
      'conversions_from_interactions_rate': ['conversions', 'interactions'],
      'cost_per_conversion': ['cost', 'conversions'],
      'cost_per_all_conversion': ['cost', 'all_conversions'],
      'all_conversion_rate': ['all_conversions', 'interactions'],
      'all_conversions_from_interactions_rate': [
        'all_conversions',
        'interactions',
      ],
    }
    if 'conversion_name' in placements.column_names:
      base_groupby = base_groupby + ['conversion_name']
      aggregation_dict.update(
        dict.fromkeys(['custom_conversions', 'custom_all_conversions'], 'sum')
      )
      relative_aggregations_dict.update(
        {
          'cost_per_custom_conversions': ['cost', 'custom_conversions'],
          'cost_per_custom_all_conversions': ['cost', 'custom_all_conversions'],
        }
      )

    aggregation_groupby = self._define_aggregation_group_by(exclusion_level)
    groupby = [
      base
      for base in base_groupby + aggregation_groupby
      if base in placements.column_names
    ]
    aggregations = {
      key: value
      for key, value in aggregation_dict.items()
      if key in placements.column_names
    }
    aggregated_placements = (
      placements.to_pandas().groupby(groupby, as_index=False).agg(aggregations)
    )
    if perform_relative_aggregations:
      for key, [numerator, denominator] in relative_aggregations_dict.items():
        if {numerator, denominator}.issubset(
          set(aggregated_placements.columns)
        ):
          aggregated_placements[key] = (
            aggregated_placements[numerator]
            / aggregated_placements[denominator]
          )
          if key == 'avg_cpm':
            aggregated_placements[key] = aggregated_placements[key] * 1000
          if key == 'ctr':
            aggregated_placements[key] = round(aggregated_placements[key], 4)
          else:
            aggregated_placements[key] = round(aggregated_placements[key], 2)
    aggregated_placements.replace([np.inf, -np.inf], 0, inplace=True)
    return gaarf.report.GaarfReport.from_pandas(aggregated_placements)

  def _define_aggregation_group_by(
    self, exclusion_level: enums.ExclusionLevelEnum
  ) -> list[str]:
    aggregation_groupby = ['account_name', 'customer_id']
    if exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
      aggregation_groupby += ['campaign_id', 'campaign_name', 'campaign_type']
    elif exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
      aggregation_groupby += [
        'campaign_id',
        'campaign_name',
        'campaign_type',
        'ad_group_id',
        'ad_group_name',
      ]
    return aggregation_groupby



from __future__ import annotations

from gaarf.api_clients import GoogleAdsApiClient
from gaarf.base_query import BaseQuery
from gaarf.query_executor import AdsReportFetcher
from gaarf.report import GaarfReport


class NegativePlacementsLists(BaseQuery):
  def __init__(self):
    self.query_text = """
            SELECT
                customer.descriptive_name AS account_name,
                customer.id AS account_id,
                shared_set.name AS name,
                shared_set.id AS id
            FROM shared_set
            WHERE shared_set.type = 'NEGATIVE_PLACEMENTS'
            AND shared_set.status = 'ENABLED'
        """


class AttachedSharedSets(BaseQuery):
  def __init__(self):
    self.query_text = """
            SELECT
                campaign.id AS campaign_id,
                shared_set.name AS name,
            FROM shared_set
            WHERE shared_set.type = 'NEGATIVE_PLACEMENTS'
            AND shared_set.status = 'ENABLED'
        """


def get_unmapped_campaigns(
  all_shared_sets: GaarfReport,
  sets_attached_to_campaigns: GaarfReport,
  ocid_mapping: dict[int, str],
) -> GaarfReport:
  ads_housekeeper_shared_sets = {}
  for shared_set in all_shared_sets:
    if shared_set.name.startswith('CPR'):
      ads_housekeeper_shared_sets[shared_set.name] = {
        'id': shared_set.id,
        'account_name': shared_set.account_name,
        'account_id': shared_set.account_id,
      }

  ads_housekeeper_campaigns_attached_to_correct_shared_sets = set()
  for shared_set in sets_attached_to_campaigns:
    if str(shared_set.campaign_id) in shared_set.name:
      ads_housekeeper_campaigns_attached_to_correct_shared_sets.add(
        shared_set.name
      )
  unmapped_campaigns = []
  for set_name, set_info in ads_housekeeper_shared_sets.items():
    if (
      set_name not in ads_housekeeper_campaigns_attached_to_correct_shared_sets
    ):
      results = [
        set_info.get('account_name'),
        set_info.get('account_id'),
        ocid_mapping.get(set_info.get('account_id'), ''),
        set_info.get('id'),
        set_name,
      ]
      unmapped_campaigns.append([results])
  return GaarfReport(
    unmapped_campaigns,
    column_names=['account_name', 'account_id', 'ocid', 'set_id', 'set_name'],
  )


def main():
  customer_ids = ['1382274117']
  report_fetcher = AdsReportFetcher(GoogleAdsApiClient())
  ocid_mapping = report_fetcher.fetch(
    'SELECT * FROM builtin.ocid_mapping', customer_ids
  )
  ocid_mapping = ocid_mapping.to_dict(
    key_column='account_id', value_column='ocid', value_column_output='scalar'
  )
  all_shared_sets = report_fetcher.fetch(
    NegativePlacementsLists(), customer_ids
  )
  sets_attached_to_campaigns = report_fetcher.fetch(
    AttachedSharedSets(), customer_ids
  )
  unmapped_campaigns = get_unmapped_campaigns(
    all_shared_sets, sets_attached_to_campaigns, ocid_mapping
  )


if __name__ == '__main__':
  main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining preview_results objects."""

# pylint: disable=C0330,
from __future__ import annotations

import dataclasses
import datetime
import enum
import uuid


class PreviewTaskStatus(enum.Enum):
  """Holds type of Task execution."""

  RUNNING = 'RUNNING'
  DONE = 'DONE'
  ERROR = 'ERROR'


@dataclasses.dataclass
class PreviewTaskAndResults:
  """Holds information of a particular preview result set.

  Attributes:
    status: execution status
    creation_time: Time when execution started.
    preview_command: The preview command
    results: Optional result details as a text field.
    id: Unique identifier of execution.
  """

  preview_command: str
  creation_time: datetime = dataclasses.field(
    default_factory=datetime.datetime.now
  )
  status: str = PreviewTaskStatus.RUNNING.value
  results: str | None = None
  id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

  DATETIME_FORMAT = '%d-%b-%Y %H:%M'

  def to_serializable_dict(self):
    """Returns a JSON-serializable dictionary representation of the object."""

    def serialize(value):
      if isinstance(value, datetime.datetime):
        return value.strftime(self.DATETIME_FORMAT)
      if isinstance(value, PreviewTaskStatus):
        return value.value
      return value

    return {
      field.name: serialize(getattr(self, field.name))
      for field in dataclasses.fields(self)
    }



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining configuration entities."""

from __future__ import annotations

import dataclasses
import uuid


@dataclasses.dataclass
class Config:
  """Represents configuration settings of the application.

  Attributes:
    mcc_id:
      Manager account perform operations.
    email_address:
      Email address to send notofications to.
    always_fetch_youtube_preview_mode:
      Whether to fetch YouTube metadata when previewing placements.
    save_to_db:
      Whether to save external parsing results to a database.
    id: Unique identifier of the configuration.
  """

  mcc_id: str
  email_address: str
  always_fetch_youtube_preview_mode: bool = True
  save_to_db: bool = True
  id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))


@dataclasses.dataclass
class CustomerIds:
  """Represents mapping between manager and child accounts.

  Attributes:
    mcc_id: Manager account perform operations.
    account_name: Child account name.
    account_name: Child account id.
  """

  mcc_id: str
  account_name: str
  id: str


@dataclasses.dataclass
class MccIds:
  """Represents mapping between manager name and its id.

  Attributes:
    id: Manager account perform operations.
    account_name: Manager account name.
  """

  id: str
  account_name: str



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining configuration entities."""

from __future__ import annotations

import dataclasses
import uuid


@dataclasses.dataclass
class Config:
  """Holds application configuration parameters.

  Attributes:
      mcc_id:
          MCC account to fetch data from.
      email_address:
          Admin email address to get notifications.
      always_fetch_youtube_preview_mode:
          Whether to force fetch YouTube data in preview mode.
      save_to_db:
          Whether to saved parsed data to DB.
      id:
          Unique config identifier.
  """

  mcc_id: str
  email_address: str
  always_fetch_youtube_preview_mode: bool = True
  save_to_db: bool = True
  id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))


@dataclasses.dataclass
class CustomerIds:
  """Contains all information on a single Google Ads account.

  Attributes:
      mcc_id: Id of an MCC account associated with a given child account.
      account_name: Name of Google Ads account.
      id: Id of Google Ads account.
  """

  mcc_id: str
  account_name: str
  id: str


@dataclasses.dataclass
class MccIds:
  """Contains all information on a single Google Ads MMC account.

  Attributes:
      id: Id of an MCC account.
      account_name: Name of Google Ads MMC account.
  """

  id: str
  account_name: str



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Holds various Enums related to the application"""

from __future__ import annotations

from enum import Enum, auto


class PlacementTypeEnum(Enum):
  WEBSITE = auto()
  YOUTUBE_VIDEO = auto()
  YOUTUBE_CHANNEL = auto()
  MOBILE_APPLICATION = auto()
  MOBILE_APP_CATEGORY = auto()


class ExclusionTypeEnum(Enum):
  GOOGLE_ADS_INFO = auto()
  YOUTUBE_CHANNEL_INFO = auto()
  YOUTUBE_VIDEO_INFO = auto()
  WEBSITE_INFO = auto()
  MOBILE_APPLICATION_INFO = auto()
  MOBILE_CATEGORY_INFO = auto()


class ExclusionLevelEnum(Enum):
  AD_GROUP = 1
  CAMPAIGN = 2
  ACCOUNT = 3
  MCC = 4


class CampaignTypeEnum(Enum):
  """Supported campaign.advertising_channel_type Enum values."""

  DISPLAY = auto()
  SEARCH = auto()
  PERFORMANCE_MAX = auto()
  DISCOVERY = auto()
  VIDEO = auto()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import abc

import sqlalchemy

from googleads_housekeeper.adapters import repository
from googleads_housekeeper.domain import events
from googleads_housekeeper.domain.core import (
  execution,
  preview_task,
  settings,
  task,
)
from googleads_housekeeper.domain.external_parsers import (
  website_parser,
  youtube_data_parser,
)
from googleads_housekeeper.domain.placement_handler import (
  entities as placement_entities,
)


class AbstractUnitOfWork(abc.ABC):
  tasks: repository.AbstractRepository
  old_tasks: repository.AbstractRepository
  settings: repository.AbstractRepository
  customer_ids: repository.AbstractRepository
  mcc_ids: repository.AbstractRepository
  website_info: repository.AbstractRepository
  youtube_channel_info: repository.AbstractRepository
  youtube_video_info: repository.AbstractRepository
  allowlisting: repository.AbstractRepository
  executions: repository.AbstractRepository
  execution_details: repository.AbstractRepository
  preview_task_and_results: repository.AbstractRepository
  published_events: list[events.Event] = []

  def __enter__(self):
    return self

  def __exit__(self, *args):
    self.rollback()

  def commit(self):
    self._commit()

  def collect_new_events(self):
    while self.published_events:
      yield self.published_events.pop(0)

  @abc.abstractmethod
  def _commit(self): ...

  @abc.abstractmethod
  def rollback(self): ...


class SqlAlchemyUnitOfWork(AbstractUnitOfWork):
  def __init__(self, session_factory: str) -> None:
    self.engine = sqlalchemy.create_engine(
      session_factory, connect_args={'check_same_thread': False}
    )
    self.session_factory = sqlalchemy.orm.sessionmaker(self.engine)
    self.published_events: list[events.Event] = []

  def __enter__(self) -> None:
    self.session = self.session_factory()
    self.tasks = repository.SqlAlchemyRepository(self.session)
    self.old_tasks = repository.SqlAlchemyRepository(
      self.session, entity=task.OldTask
    )
    self.settings = repository.SqlAlchemyRepository(
      session=self.session, entity=settings.Config
    )
    self.customer_ids = repository.SqlAlchemyRepository(
      session=self.session, entity=settings.CustomerIds
    )
    self.mcc_ids = repository.SqlAlchemyRepository(
      session=self.session, entity=settings.MccIds
    )
    self.website_info = repository.SqlAlchemyRepository(
      self.session, entity=website_parser.WebsiteInfo
    )
    self.youtube_channel_info = repository.SqlAlchemyRepository(
      self.session, entity=youtube_data_parser.ChannelInfo
    )
    self.youtube_video_info = repository.SqlAlchemyRepository(
      self.session, entity=youtube_data_parser.VideoInfo
    )
    self.allowlisting = repository.SqlAlchemyRepository(
      self.session, entity=placement_entities.AllowlistedPlacement
    )
    self.executions = repository.SqlAlchemyRepository(
      self.session, entity=execution.Execution
    )
    self.execution_details = repository.SqlAlchemyRepository(
      self.session, entity=execution.ExecutionDetails
    )
    self.preview_task_and_results = repository.SqlAlchemyRepository(
      self.session, entity=preview_task.PreviewTaskAndResults
    )
    super().__enter__()

  def __exit__(self, *args):
    super().__exit__(*args)
    self.session.close()

  def _commit(self):
    self.session.commit()

  def rollback(self):
    self.session.rollback()


class FirestoreUnitOfWork(AbstractUnitOfWork):
  def __init__(self, client: 'google.cloud.firestore.Client') -> None:
    self.client = client
    self.published_events: list[events.Event] = []

  def __enter__(self):
    self.tasks = repository.FirestoreRepository(self.client)

    self.old_tasks = repository.FirestoreRepository(
      client=self.client, entity=task.OldTask
    )
    self.settings = repository.FirestoreRepository(
      client=self.client, entity=settings.Config
    )
    self.customer_ids = repository.FirestoreRepository(
      client=self.client, entity=settings.CustomerIds
    )
    self.mcc_ids = repository.FirestoreRepository(
      client=self.client, entity=settings.MccIds
    )
    self.website_info = repository.FirestoreRepository(
      self.client, entity=website_parser.WebsiteInfo
    )
    self.youtube_channel_info = repository.FirestoreRepository(
      self.client, entity=youtube_data_parser.ChannelInfo
    )
    self.youtube_video_info = repository.FirestoreRepository(
      self.client, entity=youtube_data_parser.VideoInfo
    )
    self.allowlisting = repository.FirestoreRepository(
      self.client, entity=placement_entities.AllowlistedPlacement
    )
    self.executions = repository.FirestoreRepository(
      self.client, entity=execution.Execution
    )
    self.execution_details = repository.FirestoreRepository(
      self.client, entity=execution.ExecutionDetails
    )
    self.preview_task_and_results = repository.FirestoreRepository(
      self.client, entity=preview_task.PreviewTaskAndResults
    )
    return super().__enter__()

  def _commit(self): ...

  def rollback(self): ...



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Extracts extra info on placements from DB."""

from __future__ import annotations

import dataclasses

import gaarf

from googleads_housekeeper.adapters import repository
from googleads_housekeeper.domain.external_parsers import base_parser
from googleads_housekeeper.services import unit_of_work


class PlacementInfoExtractor:
  """Responsible for getting parsed placement information from DB.

  Attributes:
      website_info: Repository containing parsed website data.
      youtube_channel_info: Repository containing parsed YouTube channel data.
      youtube_video_info: Repository containing parsed YouTube video data.
  """

  def __init__(self, uow: unit_of_work.AbstractUnitOfWork) -> None:
    """Creates PlacementInfoExtractor repositories.

    Args:
        uow: An instance of Unit of Work.
    """
    self.uow = uow

  def extract_placement_info(
    self, placement_info: gaarf.report.GaarfRow
  ) -> dict[str, base_parser.EntityInfo]:
    """Fetches parsed information for a given placement.

    Args:
        placement_info:
            A row from report containing placement info from Google Ads.

    Returns:
        Mapping between placement parsed into type (i.e. 'website_info')
        and actual information fetched from DB.
    """
    with self.uow:
      if placement_info.placement_type == 'WEBSITE':
        return {
          'website_info': self._get_placement_from_repo(
            self.uow.website_info, placement_info.placement
          )
        }
      if placement_info.placement_type == 'YOUTUBE_CHANNEL':
        return {
          'youtube_channel_info': self._get_placement_from_repo(
            self.uow.youtube_channel_info, placement_info.placement
          )
        }

      if placement_info.placement_type == 'YOUTUBE_VIDEO':
        return {
          'youtube_video_info': self._get_placement_from_repo(
            self.uow.youtube_video_info, placement_info.placement
          )
        }
    return {}

  def _get_placement_from_repo(
    self, repo: repository.AbstractRepository, placement: str
  ) -> base_parser.EntityInfo | dict:
    """Helper method for getting data from repository.

    Args:
        repository: Repository to get data from.
        placement: Placement identifier (website_url, channel_id, video_id).

    Returns:
        Parsed placement info.
    """
    if parsed_placement_info := repo.get_by_condition('placement', placement):
      placement_type_class = type(parsed_placement_info[0])
      placement_info = dataclasses.asdict(parsed_placement_info[0])
      return placement_type_class(**placement_info)
    return {}



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining execution objects."""

from __future__ import annotations

import dataclasses
import datetime
import enum
import uuid


class ExecutionTypeEnum(enum.Enum):
  """Holds type of Task execution."""

  SCHEDULED = 0
  MANUAL = 1


@dataclasses.dataclass
class Execution:
  """Holds information of a particular task run.

  Attributes:
    task: Id of a task.
    start_date: Time when execution started.
    end_date: Time when execution ended.
    placements_excluded: Number of excluded placements.
    type: Type of execution (MANUAL, SCHEDULED).
    id: Unique identifier of execution.
  """

  task: int
  start_time: datetime.datetime
  end_time: datetime.datetime
  placements_excluded: int
  type: ExecutionTypeEnum
  id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

  def __post_init__(self) -> None:
    """Ensures that strings are always cast to proper enum."""
    self.type = self._cast_to_enum(ExecutionTypeEnum, self.type)

  def _cast_to_enum(
    self, enum_: type[enum.Enum], value: str | enum.Enum
  ) -> enum.Enum:
    """Cast to proper enum."""
    return enum_[value] if isinstance(value, str) else value


@dataclasses.dataclass
class ExecutionDetails:
  """Holds detailed information on a particular execution.

  Attributes:
    execution_id: Unique identifier of an execution.
    placement: Placement identifier that was excluded during the execution.
    placement_type: Type of excluded placement.
    reason: Reason for exclusion.
    id: Unique identifier of execution details entry.
  """

  execution_id: str
  placement: str
  placement_type: str
  reason: str
  id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
